@@ -1,23 +1,22 @@
 /*
- * Copyright (c) 2020-2023 Alex Spataru <https://github.com/alex-spataru>
+ * Serial Studio
+ * https://serial-studio.com/
  *
- * Permission is hereby granted, free of charge, to any person obtaining a copy
- * of this software and associated documentation files (the "Software"), to deal
- * in the Software without restriction, including without limitation the rights
- * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
- * copies of the Software, and to permit persons to whom the Software is
- * furnished to do so, subject to the following conditions:
+ * Copyright (C) 2020–2025 Alex Spataru
  *
- * The above copyright notice and this permission notice shall be included in
- * all copies or substantial portions of the Software.
+ * This file is dual-licensed:
  *
- * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
- * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
- * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
- * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
- * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
- * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
- * THE SOFTWARE.
+ * - Under the GNU GPLv3 (or later) for builds that exclude Pro modules.
+ * - Under the Serial Studio Commercial License for builds that include
+ *   any Pro functionality.
+ *
+ * You must comply with the terms of one of these licenses, depending
+ * on your use case.
+ *
+ * For GPL terms, see <https://www.gnu.org/licenses/gpl-3.0.html>
+ * For commercial terms, see LICENSE_COMMERCIAL.md in the project root.
+ *
+ * SPDX-License-Identifier: GPL-3.0-only OR LicenseRef-SerialStudio-Commercial
  */
 
 #include "Player.h"
@@ -27,11 +26,12 @@
 #include <QFileDialog>
 #include <QInputDialog>
 #include <QApplication>
-#include <QStandardPaths>
 
 #include "IO/Manager.h"
 #include "UI/Dashboard.h"
 #include "Misc/Utilities.h"
+#include "Misc/WorkspaceManager.h"
+#include "JSON/FrameBuilder.h"
 
 /**
  * Constructor function
@@ -40,6 +40,8 @@ CSV::Player::Player()
   : m_framePos(0)
   , m_playing(false)
   , m_timestamp("")
+  , m_startTimestampSeconds(0.0)
+  , m_useHighPrecisionTimestamps(false)
 {
   qApp->installEventFilter(this);
   connect(this, &CSV::Player::playerStateChanged, this,
@@ -66,9 +68,9 @@ bool CSV::Player::isOpen() const
 /**
  * Returns the CSV playback progress in a range from 0.0 to 1.0
  */
-qreal CSV::Player::progress() const
+double CSV::Player::progress() const
 {
-  return ((qreal)framePosition()) / frameCount();
+  return ((double)framePosition()) / frameCount();
 }
 
 /**
@@ -113,25 +115,6 @@ QString CSV::Player::filename() const
   return "";
 }
 
-/**
- * Returns the default path for CSV files
- */
-QString CSV::Player::csvFilesPath() const
-{
-  // Get file name and path
-  const auto path = QStringLiteral("%1/%2/CSV/")
-                        .arg(QStandardPaths::writableLocation(
-                                 QStandardPaths::DocumentsLocation),
-                             qApp->applicationDisplayName());
-
-  // Generate file path if required
-  QDir dir(path);
-  if (!dir.exists())
-    dir.mkpath(".");
-
-  return path;
-}
-
 /**
  * Returns the timestamp of the current data frame / row.
  */
@@ -149,6 +132,10 @@ void CSV::Player::play()
   if (m_framePos >= frameCount() - 1)
     m_framePos = 0;
 
+  // Initialize elapsed timer and record the starting CSV timestamp
+  m_startTimestamp = getDateTime(m_framePos);
+  m_elapsedTimer.start();
+
   m_playing = true;
   Q_EMIT playerStateChanged();
 }
@@ -179,14 +166,22 @@ void CSV::Player::toggle()
  */
 void CSV::Player::openFile()
 {
-  // Get file name
-  auto file = QFileDialog::getOpenFileName(
-      nullptr, tr("Select CSV file"), csvFilesPath(),
-      tr("CSV files") + QStringLiteral(" (*.csv)"));
+  auto *dialog = new QFileDialog(nullptr, tr("Select CSV file"),
+                                 Misc::WorkspaceManager::instance().path("CSV"),
+                                 tr("CSV files (*.csv)"));
 
-  // Open CSV file
-  if (!file.isEmpty())
-    openFile(file);
+  dialog->setFileMode(QFileDialog::ExistingFile);
+  dialog->setOption(QFileDialog::DontUseNativeDialog);
+
+  connect(dialog, &QFileDialog::fileSelected, this,
+          [this, dialog](const QString &path) {
+            if (!path.isEmpty())
+              openFile(path);
+
+            dialog->deleteLater();
+          });
+
+  dialog->open();
 }
 
 /**
@@ -201,6 +196,11 @@ void CSV::Player::closeFile()
   m_csvData.squeeze();
   m_playing = false;
   m_timestamp = "--.--";
+  m_timestampCache.clear();
+  m_useHighPrecisionTimestamps = false;
+  m_startTimestampSeconds = 0.0;
+
+  JSON::FrameBuilder::instance().registerQuickPlotHeaders(QStringList());
 
   Q_EMIT openChanged();
   Q_EMIT timestampChanged();
@@ -211,8 +211,8 @@ void CSV::Player::closeFile()
  * @brief Reads & processes the next CSV row, capped at the last row.
  *
  * Moves the frame position forward by one, up to the last frame in the CSV.
- * Resets and repopulates the dashboard with the appropriate frames for
- * synchronized display.
+ * Clears plot data and repopulates the dashboard with the appropriate frames
+ * for synchronized display.
  */
 void CSV::Player::nextFrame()
 {
@@ -221,12 +221,13 @@ void CSV::Player::nextFrame()
     // Increase the frame position
     ++m_framePos;
 
+    // Clear only plot data (preserves widget layout)
+    UI::Dashboard::instance().clearPlotData();
+
     // Populate the dashboard with a range of frames up to the new position
-    UI::Dashboard::instance().resetData(false);
     int framesToLoad = UI::Dashboard::instance().points();
     int startFrame = std::max(1, m_framePos - framesToLoad);
-    for (int i = startFrame; i <= m_framePos; ++i)
-      IO::Manager::instance().processPayload(getFrame(i));
+    processFrameBatch(startFrame, m_framePos);
 
     // Keep timestamp and data in sync
     updateData();
@@ -237,8 +238,8 @@ void CSV::Player::nextFrame()
  * @brief Reads & processes the previous CSV row, capped at the first row.
  *
  * Moves the frame position backward by one, down to the first frame in the CSV.
- * Resets and repopulates the dashboard with the appropriate frames for
- * synchronized display.
+ * Clears plot data and repopulates the dashboard with the appropriate frames
+ * for synchronized display.
  */
 void CSV::Player::previousFrame()
 {
@@ -247,12 +248,13 @@ void CSV::Player::previousFrame()
     // Decrease the frame position
     --m_framePos;
 
+    // Clear only plot data (preserves widget layout)
+    UI::Dashboard::instance().clearPlotData();
+
     // Populate the dashboard with a range of frames up to the new position
-    UI::Dashboard::instance().resetData(false);
     int framesToLoad = UI::Dashboard::instance().points();
     int startFrame = std::max(1, m_framePos - framesToLoad);
-    for (int i = startFrame; i <= m_framePos; ++i)
-      IO::Manager::instance().processPayload(getFrame(i));
+    processFrameBatch(startFrame, m_framePos);
 
     // Keep timestamp and data in sync
     updateData();
@@ -263,7 +265,7 @@ void CSV::Player::previousFrame()
  * @brief Opens a CSV file, processes its data, and prepares it for playback.
  *
  * This function attempts to open the specified CSV file for reading and
- * processes the data for replaying. It checks if a device is connected and,
+ * processes the data for replaying. It checks if a device is isConnected and,
  * if so, asks the user to disconnect it.
  *
  * The function reads the CSV file into a string matrix (`m_csvData`), validates
@@ -291,13 +293,13 @@ void CSV::Player::openFile(const QString &filePath)
   closeFile();
 
   // Device is connected, warn user & disconnect
-  if (IO::Manager::instance().connected())
+  if (IO::Manager::instance().isConnected())
   {
     auto response = Misc::Utilities::showMessageBox(
         tr("Device Connection Active"),
         tr("To use this feature, you must disconnect from the device. "
            "Do you want to proceed?"),
-        qAppName(), QMessageBox::No | QMessageBox::Yes);
+        QMessageBox::Warning, qAppName(), QMessageBox::No | QMessageBox::Yes);
     if (response == QMessageBox::Yes)
       IO::Manager::instance().disconnectDevice();
     else
@@ -333,9 +335,42 @@ void CSV::Player::openFile(const QString &filePath)
         m_csvData.append(row);
     }
 
-    // Validate the cell at (1,1) for date/time format
-    if (!getDateTime(1).isValid())
+    // Detect timestamp format and validate
+    // First try high-precision format (fractional seconds)
+    bool error = false;
+    QString firstCell = getCellValue(1, 0, error);
+    double firstTimestamp = error ? -1.0 : getTimestampSeconds(firstCell);
+
+    if (firstTimestamp >= 0.0)
     {
+      // High-precision timestamps detected - build cache
+      m_timestampCache.clear();
+      m_timestampCache.reserve(m_csvData.count());
+
+      // Cache all timestamps for efficient playback
+      for (int i = 0; i < m_csvData.count(); ++i)
+      {
+        bool err = false;
+        QString cell = getCellValue(i, 0, err);
+        double ts = err ? 0.0 : getTimestampSeconds(cell);
+        m_timestampCache.append(ts);
+      }
+
+      m_startTimestampSeconds = m_timestampCache[1];
+      m_useHighPrecisionTimestamps = true;
+    }
+    // Fall back to legacy date/time format
+    else if (getDateTime(1).isValid())
+    {
+      m_useHighPrecisionTimestamps = false;
+      m_timestampCache.clear();
+    }
+    // Neither format worked - prompt user
+    else
+    {
+      m_useHighPrecisionTimestamps = false;
+      m_timestampCache.clear();
+
       // Ask user to select date/time column or set interval manually
       if (!promptUserForDateTimeOrInterval())
       {
@@ -344,10 +379,21 @@ void CSV::Player::openFile(const QString &filePath)
       }
     }
 
+    // Send the header row before removing it
+    sendHeaderFrame();
+
     // Remove the header row from the data
     m_framePos = 0;
     m_csvData.removeFirst();
 
+    // Adjust cached timestamps if header was removed
+    if (m_useHighPrecisionTimestamps && !m_timestampCache.isEmpty())
+    {
+      m_timestampCache.removeFirst();
+      if (!m_timestampCache.isEmpty())
+        m_startTimestampSeconds = m_timestampCache[0];
+    }
+
     // Begin reading data
     if (m_csvData.count() >= 2)
     {
@@ -361,7 +407,8 @@ void CSV::Player::openFile(const QString &filePath)
       Misc::Utilities::showMessageBox(
           tr("Insufficient Data in CSV File"),
           tr("The CSV file must contain at least two frames (data rows) to "
-             "proceed. Please check the file and try again."));
+             "proceed. Please check the file and try again."),
+          QMessageBox::Critical);
       closeFile();
     }
   }
@@ -371,7 +418,7 @@ void CSV::Player::openFile(const QString &filePath)
   {
     Misc::Utilities::showMessageBox(
         tr("Cannot read CSV file"),
-        tr("Please check file permissions & location"));
+        tr("Please check file permissions & location"), QMessageBox::Critical);
     closeFile();
   }
 }
@@ -411,7 +458,7 @@ void CSV::Player::openFile(const QString &filePath)
  * @note A silent reset is performed to refresh the dashboard plots without
  *       emitting unnecessary signals.
  */
-void CSV::Player::setProgress(const qreal progress)
+void CSV::Player::setProgress(const double progress)
 {
   // Clamp progress between 0 and 1
   const auto validProgress = std::clamp(progress, 0.0, 1.0);
@@ -426,9 +473,11 @@ void CSV::Player::setProgress(const qreal progress)
   // Only process if position changes
   if (newFramePos != m_framePos)
   {
-    // Update frame position and reset dashboard data silently
+    // Update frame position
     m_framePos = newFramePos;
-    UI::Dashboard::instance().resetData(false);
+
+    // Clear only plot data (preserves widget layout)
+    UI::Dashboard::instance().clearPlotData();
 
     // Calculate frames to load around the new frame position
     int framesToLoad = UI::Dashboard::instance().points();
@@ -436,8 +485,7 @@ void CSV::Player::setProgress(const qreal progress)
     int endFrame = std::min(frameCount() - 1, m_framePos);
 
     // Populate dashboard with frames within capped range
-    for (int i = startFrame; i <= endFrame; ++i)
-      IO::Manager::instance().processPayload(getFrame(i));
+    processFrameBatch(startFrame, endFrame);
 
     // Update with current data
     updateData();
@@ -445,69 +493,176 @@ void CSV::Player::setProgress(const qreal progress)
 }
 
 /**
- * Generates a JSON data frame by combining the values of the current CSV
- * row & the structure of the JSON map file loaded in the @c JsonParser class.
- *
- * If playback is enabled, this function calculates the difference in
- * milliseconds between the current row and the next row & schedules a re-call
- * of this function using a timer.
+ * @brief Updates the timestamp display for the current frame position.
+ */
+void CSV::Player::updateTimestampDisplay()
+{
+  bool err = true;
+  auto ts = getCellValue(m_framePos, 0, err);
+  if (err)
+    return;
+
+  double seconds = getTimestampSeconds(ts);
+  if (seconds >= 0.0)
+    m_timestamp = formatTimestamp(seconds);
+
+  else
+  {
+    auto frameTime = getDateTime(ts);
+    if (frameTime.isValid() && m_startTimestamp.isValid())
+    {
+      qint64 elapsedMs = m_startTimestamp.msecsTo(frameTime);
+      m_timestamp = formatTimestamp(elapsedMs / 1000.0);
+    }
+
+    else
+      m_timestamp = ts;
+  }
+
+  Q_EMIT timestampChanged();
+}
+
+/**
+ * @brief Processes current frame and schedules next frame for playback.
  */
 void CSV::Player::updateData()
 {
-  // File not open, abort
   if (!isOpen())
     return;
 
-  // Update timestamp string
-  bool error = true;
-  auto timestamp = getCellValue(framePosition(), 0, error);
-  if (!error)
+  IO::Manager::instance().processPayload(getFrame(framePosition()));
+  updateTimestampDisplay();
+
+  if (!isPlaying())
+    return;
+
+  if (framePosition() >= frameCount() - 1)
   {
-    m_timestamp = timestamp;
-    IO::Manager::instance().processPayload(getFrame(framePosition()));
-    Q_EMIT timestampChanged();
+    pause();
+    return;
   }
 
-  // If the user wants to 'play' the CSV, get time difference between this
-  // frame and the next frame & schedule an automated update
-  if (!error && isPlaying())
+  const qint64 elapsedMs = m_elapsedTimer.elapsed();
+  qint64 msUntilNext = 0;
+
+  if (m_useHighPrecisionTimestamps)
   {
-    // Get first frame
-    if (framePosition() < frameCount() - 1)
+    if (framePosition() + 1 >= m_timestampCache.size())
     {
-      // Obtain time for current & next frame
-      auto currTime = getDateTime(framePosition());
-      auto nextTime = getDateTime(framePosition() + 1);
+      pause();
+      return;
+    }
 
-      // No error, calculate difference & schedule update
-      if (currTime.isValid() && nextTime.isValid())
+    const double targetTime = m_startTimestampSeconds + (elapsedMs / 1000.0);
+    const double nextTime = m_timestampCache[framePosition() + 1];
+    msUntilNext
+        = qMax(0LL, static_cast<qint64>((nextTime - targetTime) * 1000.0));
+  }
+
+  else
+  {
+    const QDateTime targetTime = m_startTimestamp.addMSecs(elapsedMs);
+    const auto nextTime = getDateTime(framePosition() + 1);
+    if (!nextTime.isValid())
+    {
+      pause();
+      return;
+    }
+
+    msUntilNext = targetTime.msecsTo(nextTime);
+  }
+
+  if (msUntilNext <= 0)
+  {
+    constexpr int kMaxBatchSize = 100;
+    int processed = 0;
+    while (m_framePos < frameCount() - 1 && processed < kMaxBatchSize
+           && msUntilNext <= 0)
+    {
+      ++m_framePos;
+      IO::Manager::instance().processPayload(getFrame(m_framePos));
+      ++processed;
+
+      if (m_useHighPrecisionTimestamps)
       {
-        // Obtain millis between the two frames
-        const auto msecsToNextF = abs(currTime.msecsTo(nextTime));
-
-        // Jump to next frame
-        QTimer::singleShot(msecsToNextF, Qt::PreciseTimer, this, [=] {
-          if (isOpen() && isPlaying() && framePosition() < frameCount())
-          {
-            ++m_framePos;
-            updateData();
-          }
-        });
+        if (m_framePos < m_timestampCache.size())
+        {
+          const double target
+              = m_startTimestampSeconds + (m_elapsedTimer.elapsed() / 1000.0);
+          const double next = m_timestampCache[m_framePos];
+          msUntilNext
+              = qMax(0LL, static_cast<qint64>((next - target) * 1000.0));
+        }
       }
 
-      // Error - pause playback
       else
       {
-        pause();
-        qWarning() << "Error getting timestamp difference" << currTime
-                   << nextTime;
+        auto target = m_startTimestamp.addMSecs(m_elapsedTimer.elapsed());
+        auto next = getDateTime(m_framePos);
+        if (next.isValid())
+          msUntilNext = target.msecsTo(next);
       }
     }
 
-    // Pause at end of CSV
+    updateTimestampDisplay();
+
+    if (m_framePos < frameCount() - 1)
+      QTimer::singleShot(qMax(0LL, msUntilNext), Qt::PreciseTimer, this,
+                         &CSV::Player::updateData);
     else
       pause();
   }
+
+  else
+  {
+    QTimer::singleShot(msUntilNext, Qt::PreciseTimer, this, [this] {
+      if (!isOpen() || !isPlaying())
+        return;
+
+      ++m_framePos;
+      updateData();
+    });
+  }
+}
+
+/**
+ * @brief Processes a batch of frames synchronously for efficient scrollback.
+ *
+ * This function processes frames from startFrame to endFrame (inclusive)
+ * without using timers, which is more efficient for scrollback operations
+ * where we need to rebuild plot history.
+ *
+ * @param startFrame The first frame index to process.
+ * @param endFrame The last frame index to process (inclusive).
+ */
+void CSV::Player::processFrameBatch(int startFrame, int endFrame)
+{
+  for (int i = startFrame; i <= endFrame; ++i)
+    IO::Manager::instance().processPayload(getFrame(i));
+}
+
+/**
+ * @brief Registers CSV headers with Quick Plot
+ *
+ * Extracts column names from the CSV header row and explicitly registers
+ * them with the FrameBuilder. Skips the first column (timestamp) as it
+ * is not used for plotting. This should be called before the header row
+ * is removed from m_csvData.
+ */
+void CSV::Player::sendHeaderFrame()
+{
+  if (m_csvData.isEmpty())
+    return;
+
+  const auto &headerRow = m_csvData.first();
+  if (headerRow.size() <= 1)
+    return;
+
+  QStringList headers;
+  for (int i = 1; i < headerRow.size(); ++i)
+    headers.append(headerRow[i]);
+
+  JSON::FrameBuilder::instance().registerQuickPlotHeaders(headers);
 }
 
 /**
@@ -529,7 +684,8 @@ bool CSV::Player::promptUserForDateTimeOrInterval()
   {
     Misc::Utilities::showMessageBox(
         tr("Invalid CSV"),
-        tr("The CSV file does not contain any data or headers."));
+        tr("The CSV file does not contain any data or headers."),
+        QMessageBox::Critical);
     return false;
   }
 
@@ -577,8 +733,9 @@ bool CSV::Player::promptUserForDateTimeOrInterval()
       int columnIndex = headerLabels.indexOf(column);
       if (columnIndex == -1)
       {
-        Misc::Utilities::showMessageBox(
-            tr("Invalid Selection"), tr("The selected column is not valid."));
+        Misc::Utilities::showMessageBox(tr("Invalid Selection"),
+                                        tr("The selected column is not valid."),
+                                        QMessageBox::Critical);
         return false;
       }
 
@@ -689,6 +846,13 @@ QDateTime CSV::Player::getDateTime(const int row)
  */
 QDateTime CSV::Player::getDateTime(const QString &cell)
 {
+  // If the cell is a pure number (fractional seconds), don't try to parse as
+  // date
+  bool isNumber = false;
+  cell.toDouble(&isNumber);
+  if (isNumber)
+    return QDateTime(); // Return invalid QDateTime for fractional seconds
+
   // Initialize parameters
   QDateTime dateTime;
 
@@ -711,6 +875,66 @@ QDateTime CSV::Player::getDateTime(const QString &cell)
   return dateTime;
 }
 
+/**
+ * @brief Parses a timestamp in fractional seconds format from the given @a row.
+ *
+ * @param row The row index to parse the timestamp from.
+ * @return The timestamp in seconds, or -1.0 if invalid.
+ */
+double CSV::Player::getTimestampSeconds(int row)
+{
+  // Use cached value if available and valid
+  if (m_useHighPrecisionTimestamps && row >= 0 && row < m_timestampCache.size())
+    return m_timestampCache[row];
+
+  // Fallback to parsing from cell
+  bool error = false;
+  QString cell = getCellValue(row, 0, error);
+  if (error)
+    return -1.0;
+
+  return getTimestampSeconds(cell);
+}
+
+/**
+ * @brief Parses a timestamp in fractional seconds format from the given @a
+ * cell.
+ *
+ * This method attempts to parse the cell as a decimal number representing
+ * seconds since Unix epoch with nanosecond precision.
+ *
+ * @param cell The cell string to parse.
+ * @return The timestamp in seconds, or -1.0 if invalid.
+ */
+double CSV::Player::getTimestampSeconds(const QString &cell)
+{
+  bool ok = false;
+  double timestamp = cell.toDouble(&ok);
+
+  if (ok && timestamp >= 0.0)
+    return timestamp;
+
+  return -1.0;
+}
+
+/**
+ * @brief Formats a timestamp in fractional seconds to HH:MM:SS.mmm format.
+ *
+ * @param seconds The timestamp in seconds.
+ * @return Formatted timestamp string.
+ */
+QString CSV::Player::formatTimestamp(double seconds) const
+{
+  int hours = static_cast<int>(seconds / 3600.0);
+  int minutes = static_cast<int>((seconds - hours * 3600.0) / 60.0);
+  double secs = seconds - hours * 3600.0 - minutes * 60.0;
+
+  return QString("%1:%2:%3")
+      .arg(hours, 2, 10, QChar('0'))
+      .arg(minutes, 2, 10, QChar('0'))
+      .arg(secs, 6, 'f', 3, QChar('0'));
+}
+
 /**
  * Generates a frame from the data at the given @a row. The first item of each
  * row is ignored because it contains the RX date/time, which is used to
@@ -741,10 +965,10 @@ QByteArray CSV::Player::getFrame(const int row)
  * error occurs or the cell does not exist, the value of @a error shall be set
  * to @c true.
  */
-const QString &CSV::Player::getCellValue(const int row, const int column,
-                                         bool &error)
+const QString CSV::Player::getCellValue(const int row, const int column,
+                                        bool &error)
 {
-  static auto defaultValue = QStringLiteral("");
+  static auto defaultValue = QLatin1String("");
 
   if (m_csvData.count() > row)
   {
