@@ -1,23 +1,22 @@
 /*
- * Copyright (c) 2020-2023 Alex Spataru <https://github.com/alex-spataru>
+ * Serial Studio
+ * https://serial-studio.com/
  *
- * Permission is hereby granted, free of charge, to any person obtaining a copy
- * of this software and associated documentation files (the "Software"), to deal
- * in the Software without restriction, including without limitation the rights
- * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
- * copies of the Software, and to permit persons to whom the Software is
- * furnished to do so, subject to the following conditions:
+ * Copyright (C) 2020–2025 Alex Spataru
  *
- * The above copyright notice and this permission notice shall be included in
- * all copies or substantial portions of the Software.
+ * This file is dual-licensed:
  *
- * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
- * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
- * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
- * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
- * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
- * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
- * THE SOFTWARE.
+ * - Under the GNU GPLv3 (or later) for builds that exclude Pro modules.
+ * - Under the Serial Studio Commercial License for builds that include
+ *   any Pro functionality.
+ *
+ * You must comply with the terms of one of these licenses, depending
+ * on your use case.
+ *
+ * For GPL terms, see <https://www.gnu.org/licenses/gpl-3.0.html>
+ * For commercial terms, see LICENSE_COMMERCIAL.md in the project root.
+ *
+ * SPDX-License-Identifier: GPL-3.0-only OR LicenseRef-SerialStudio-Commercial
  */
 
 #include <QOperatingSystemVersion>
@@ -35,8 +34,10 @@
 IO::Drivers::BluetoothLE::BluetoothLE()
   : m_deviceIndex(-1)
   , m_deviceConnected(false)
+  , m_adapterAvailable(false)
   , m_service(nullptr)
   , m_controller(nullptr)
+  , m_localDevice(nullptr)
   , m_discoveryAgent(nullptr)
 {
   connect(this, &IO::Drivers::BluetoothLE::deviceIndexChanged, this,
@@ -44,8 +45,8 @@ IO::Drivers::BluetoothLE::BluetoothLE()
 
   connect(this, &IO::Drivers::BluetoothLE::error, this,
           [=](const QString &message) {
-            Misc::Utilities::showMessageBox(tr("BLE I/O Module Error"),
-                                            message);
+            Misc::Utilities::showMessageBox(tr("BLE I/O Module Error"), message,
+                                            QMessageBox::Critical);
           });
 }
 
@@ -140,7 +141,7 @@ bool IO::Drivers::BluetoothLE::isWritable() const
  */
 bool IO::Drivers::BluetoothLE::configurationOk() const
 {
-  return operatingSystemSupported() && deviceIndex() >= 0;
+  return operatingSystemSupported() && adapterAvailable() && deviceIndex() >= 0;
 }
 
 /**
@@ -166,16 +167,16 @@ quint64 IO::Drivers::BluetoothLE::write(const QByteArray &data)
 
     else
     {
-      qWarning()
-          << "Failed to write data to BLE device: invalid characteristic";
-      return -1;
+      qWarning() << "Failed to write to BLE device: invalid characteristic";
+
+      return 0;
     }
   }
 
   qWarning() << "Failed to write data to BLE device: ensure that a "
                 "characteristic is selected";
 
-  return -1;
+  return 0;
 }
 
 /**
@@ -243,6 +244,16 @@ bool IO::Drivers::BluetoothLE::operatingSystemSupported() const
   return true;
 }
 
+/**
+ * Returns true if a Bluetooth adapter is available on the system.
+ * This checks if a valid Bluetooth adapter exists, regardless of whether
+ * it is powered on or off.
+ */
+bool IO::Drivers::BluetoothLE::adapterAvailable() const
+{
+  return m_adapterAvailable;
+}
+
 //------------------------------------------------------------------------------
 // Driver specifics
 //------------------------------------------------------------------------------
@@ -314,6 +325,13 @@ void IO::Drivers::BluetoothLE::startDiscovery()
   if (!operatingSystemSupported())
     return;
 
+  // Initialize adapter on first use
+  initializeBluetoothAdapter();
+
+  // Check if adapter is available
+  if (!adapterAvailable())
+    return;
+
   // Close previous device
   close();
 
@@ -406,6 +424,10 @@ void IO::Drivers::BluetoothLE::selectService(const int index)
   // Ensure that index is valid
   if (index >= 1 && index <= m_serviceNames.count())
   {
+    // Additional bounds check against actual services list
+    if (!m_controller || index - 1 >= m_controller->services().count())
+      return;
+
     // Generate service handler & connect signals/slots
     auto serviceUuid = m_controller->services().at(index - 1);
     m_service = m_controller->createServiceObject(serviceUuid, this);
@@ -456,7 +478,8 @@ void IO::Drivers::BluetoothLE::setCharacteristicIndex(const int index)
     m_selectedCharacteristic = -1;
 
   // Query characteristic for information
-  if (m_selectedCharacteristic >= 0)
+  if (m_selectedCharacteristic >= 0
+      && m_selectedCharacteristic < m_characteristics.count())
   {
     // Obtain the characteristic
     const auto &c = m_characteristics.at(m_selectedCharacteristic);
@@ -469,7 +492,7 @@ void IO::Drivers::BluetoothLE::setCharacteristicIndex(const int index)
 
     // Display current value
     if (!c.value().isEmpty())
-      processData(c.value());
+      Q_EMIT dataReceived(c.value());
   }
 
   // Update UI
@@ -598,9 +621,6 @@ void IO::Drivers::BluetoothLE::onDiscoveryError(
 {
   switch (e)
   {
-    case QBluetoothDeviceDiscoveryAgent::PoweredOffError:
-      Q_EMIT error(tr("Bluetooth adapter is off!"));
-      break;
     case QBluetoothDeviceDiscoveryAgent::InvalidBluetoothAdapterError:
       Q_EMIT error(tr("Invalid Bluetooth adapter!"));
       break;
@@ -638,5 +658,65 @@ void IO::Drivers::BluetoothLE::onCharacteristicChanged(
   const bool anyCharacteristic = (m_selectedCharacteristic == -1);
   const bool current = (info == m_characteristics.at(m_selectedCharacteristic));
   if (anyCharacteristic || current)
-    processData(value);
+    Q_EMIT dataReceived(value);
+}
+
+/**
+ * Handles changes to the Bluetooth adapter's host mode state.
+ * This is called when the adapter is powered on/off or physically
+ * connected/disconnected (e.g., USB Bluetooth dongles).
+ */
+void IO::Drivers::BluetoothLE::onHostModeStateChanged(
+    QBluetoothLocalDevice::HostMode state)
+{
+  bool wasAvailable = m_adapterAvailable;
+  m_adapterAvailable = (state != QBluetoothLocalDevice::HostPoweredOff);
+
+  if (wasAvailable != m_adapterAvailable)
+  {
+    Q_EMIT adapterAvailabilityChanged();
+
+    if (!m_adapterAvailable)
+    {
+      // If adapter becomes unavailable while scanning, stop scanning
+      if (m_discoveryAgent && m_discoveryAgent->isActive())
+        m_discoveryAgent->stop();
+
+      // Clear any discovered devices
+      m_devices.clear();
+      m_deviceIndex = -1;
+      m_deviceNames.clear();
+      Q_EMIT devicesChanged();
+    }
+    else
+    {
+      // Adapter became available, automatically start discovery
+      startDiscovery();
+    }
+  }
+}
+
+/**
+ * Initializes the Bluetooth adapter detection.
+ * This is called lazily when the Bluetooth driver is first used.
+ */
+void IO::Drivers::BluetoothLE::initializeBluetoothAdapter()
+{
+  // Already initialized or OS not supported
+  if (m_localDevice || !operatingSystemSupported())
+    return;
+
+  // Create local device instance
+  m_localDevice = new QBluetoothLocalDevice(this);
+
+  // Check initial adapter state
+  auto hostMode = m_localDevice->hostMode();
+  m_adapterAvailable = (hostMode != QBluetoothLocalDevice::HostPoweredOff);
+
+  // Monitor adapter state changes
+  connect(m_localDevice, &QBluetoothLocalDevice::hostModeStateChanged, this,
+          &IO::Drivers::BluetoothLE::onHostModeStateChanged);
+
+  // Notify QML of initial state
+  Q_EMIT adapterAvailabilityChanged();
 }
