@@ -1,23 +1,22 @@
 /*
- * Copyright (c) 2020-2023 Alex Spataru <https://github.com/alex-spataru>
+ * Serial Studio
+ * https://serial-studio.com/
  *
- * Permission is hereby granted, free of charge, to any person obtaining a copy
- * of this software and associated documentation files (the "Software"), to deal
- * in the Software without restriction, including without limitation the rights
- * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
- * copies of the Software, and to permit persons to whom the Software is
- * furnished to do so, subject to the following conditions:
+ * Copyright (C) 2020–2025 Alex Spataru
  *
- * The above copyright notice and this permission notice shall be included in
- * all copies or substantial portions of the Software.
+ * This file is dual-licensed:
  *
- * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
- * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
- * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
- * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
- * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
- * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
- * THE SOFTWARE.
+ * - Under the GNU GPLv3 (or later) for builds that exclude Pro modules.
+ * - Under the Serial Studio Commercial License for builds that include
+ *   any Pro functionality.
+ *
+ * You must comply with the terms of one of these licenses, depending
+ * on your use case.
+ *
+ * For GPL terms, see <https://www.gnu.org/licenses/gpl-3.0.html>
+ * For commercial terms, see LICENSE_COMMERCIAL.md in the project root.
+ *
+ * SPDX-License-Identifier: GPL-3.0-only OR LicenseRef-SerialStudio-Commercial
  */
 
 #include <QPainter>
@@ -25,7 +24,7 @@
 #include <QFontMetrics>
 #include <QApplication>
 
-#include "IO/Console.h"
+#include "Console/Handler.h"
 #include "IO/Manager.h"
 #include "Misc/Translator.h"
 #include "Misc/TimerEvents.h"
@@ -33,6 +32,15 @@
 #include "Misc/ThemeManager.h"
 #include "UI/Widgets/Terminal.h"
 
+#ifdef BUILD_COMMERCIAL
+#  include "Licensing/LemonSqueezy.h"
+#endif
+
+/**
+ * @brief Define the number of max lines supported
+ */
+constexpr int MAX_LINES = 1000;
+
 /**
  * @brief Constructs a Terminal object with the given parent item.
  *
@@ -52,7 +60,7 @@
  * - Configures the initial color palette by calling `onThemeChanged()`.
  * - Connects to the theme manager to update the palette whenever the theme
  *   changes.
- * - Connects to the IO Console handler to receive and append new data for
+ * - Connects to the IO Handler handler to receive and append new data for
  *   display.
  * - Clears the screen when a device connection status changes.
  * - Sets up a cursor blink timer and connects it to toggle the cursor
@@ -86,10 +94,12 @@ Widgets::Terminal::Terminal(QQuickItem *parent)
   setAcceptedMouseButtons(Qt::AllButtons);
 
   // Set rendering hints
-  setMipmap(false);
-  setAntialiasing(true);
+  setMipmap(true);
   setOpaquePainting(true);
 
+  // Load the welcome guide
+  loadWelcomeGuide();
+
   // Set font
   setFont(Misc::CommonFonts::instance().monoFont());
 
@@ -98,25 +108,38 @@ Widgets::Terminal::Terminal(QQuickItem *parent)
   connect(&Misc::ThemeManager::instance(), &Misc::ThemeManager::themeChanged,
           this, &Widgets::Terminal::onThemeChanged);
 
-  // Receive data from the IO::Console handler
-  connect(&IO::Console::instance(), &IO::Console::displayString, this,
-          &Widgets::Terminal::append, Qt::QueuedConnection);
+  // Receive data from the Console::Console handler
+  connect(&Console::Handler::instance(), &Console::Handler::displayString, this,
+          &Widgets::Terminal::append);
 
   // Clear the screen when device is connected/disconnected
-  connect(&IO::Manager::instance(), &IO::Manager::connectedChanged, this, [=] {
-    if (IO::Manager::instance().connected())
-      clear();
-  });
+  connect(&IO::Manager::instance(), &IO::Manager::connectedChanged, this,
+          [=, this] {
+            if (IO::Manager::instance().isConnected())
+              clear();
+            else if (m_data.isEmpty())
+              loadWelcomeGuide();
+          });
 
   // Redraw widget as soon as it is visible
-  connect(this, &Widgets::Terminal::visibleChanged, this, [=] {
+  connect(this, &Widgets::Terminal::visibleChanged, this, [=, this] {
     if (isVisible())
       update();
   });
 
+  // Update welcome guide when Serial Studio changes its activation status
+#ifdef BUILD_COMMERCIAL
+  connect(&Licensing::LemonSqueezy::instance(),
+          &Licensing::LemonSqueezy::activatedChanged, this,
+          &Widgets::Terminal::loadWelcomeGuide);
+#endif
+
   // Change character widths when changing language
   connect(&Misc::Translator::instance(), &Misc::Translator::languageChanged,
-          this, [=] { setFont(Misc::CommonFonts::instance().monoFont()); });
+          this, [=, this] {
+            loadWelcomeGuide();
+            setFont(Misc::CommonFonts::instance().monoFont());
+          });
 
   // Blink the cursor
   m_cursorTimer.start(200);
@@ -126,8 +149,8 @@ Widgets::Terminal::Terminal(QQuickItem *parent)
 
   // Redraw the widget only when necessary
   m_stateChanged = true;
-  connect(&Misc::TimerEvents::instance(), &Misc::TimerEvents::timeout24Hz, this,
-          [=] {
+  connect(&Misc::TimerEvents::instance(), &Misc::TimerEvents::uiTimeout, this,
+          [=, this] {
             if (isVisible() && m_stateChanged)
             {
               m_stateChanged = false;
@@ -347,9 +370,9 @@ void Widgets::Terminal::paint(QPainter *painter)
 
     // Set scrollbar position
     int x = width() - scrollbarWidth - m_borderX;
-    int y = (m_scrollOffsetY / static_cast<float>(lineCount() - linesPerPage()))
-                * (availableHeight - scrollbarHeight)
-            - m_borderY;
+    y = (m_scrollOffsetY / static_cast<float>(lineCount() - linesPerPage()))
+            * (availableHeight - scrollbarHeight)
+        - m_borderY;
     y = qMax(m_borderY, y);
 
     // Draw the scrollbar
@@ -386,7 +409,7 @@ int Widgets::Terminal::charHeight() const
  *
  * @return The QFont object representing the terminal's current font.
  */
-QFont Widgets::Terminal::font() const
+const QFont &Widgets::Terminal::font() const
 {
   return m_font;
 }
@@ -396,7 +419,7 @@ QFont Widgets::Terminal::font() const
  *
  * @return The QPalette object representing the terminal's color palette.
  */
-QPalette Widgets::Terminal::palette() const
+const QPalette &Widgets::Terminal::palette() const
 {
   return m_palette;
 }
@@ -503,7 +526,7 @@ int Widgets::Terminal::maxCharsPerLine() const
  * @return A QPoint representing the current cursor position, in character
  *         coordinates.
  */
-QPoint Widgets::Terminal::cursorPosition() const
+const QPoint &Widgets::Terminal::cursorPosition() const
 {
   return m_cursorPosition;
 }
@@ -810,6 +833,18 @@ void Widgets::Terminal::onThemeChanged()
   // clang-format on
 }
 
+/**
+ * @brief Displays the localized welcome guide in the terminal widget, without
+ *        modifying the terminal output.
+ */
+void Widgets::Terminal::loadWelcomeGuide()
+{
+  clear();
+  setAutoscroll(false);
+  append(Misc::Translator::instance().welcomeConsoleText());
+  setAutoscroll(true);
+}
+
 /**
  * @brief Appends a string of data to the terminal, processing each character
  *        accordingly.
@@ -825,7 +860,7 @@ void Widgets::Terminal::onThemeChanged()
  * @see processText(), processEscape(), processFormat(), processResetFont(),
  * appendString()
  */
-void Widgets::Terminal::append(const QString &data)
+void Widgets::Terminal::append(QStringView data)
 {
   QString text;
   auto it = data.constBegin();
@@ -873,10 +908,27 @@ void Widgets::Terminal::append(const QString &data)
  *
  * @see replaceData(), setCursorPosition(), autoscroll()
  */
-void Widgets::Terminal::appendString(const QString &string)
+void Widgets::Terminal::appendString(QStringView string)
 {
+  // Ensure buffer memory does not exceed MAX_LINES
+  const int linesToDrop = m_data.size() - MAX_LINES + 1;
+  if (m_data.size() >= MAX_LINES && linesToDrop > 0)
+  {
+    m_data.erase(m_data.begin(), m_data.begin() + linesToDrop);
+
+    if (m_cursorPosition.y() >= linesToDrop)
+      m_cursorPosition.setY(m_cursorPosition.y() - linesToDrop);
+    else
+      m_cursorPosition.setY(0);
+
+    if (m_scrollOffsetY >= linesToDrop)
+      m_scrollOffsetY -= linesToDrop;
+    else
+      m_scrollOffsetY = 0;
+  }
+
   // Register each character in the provided string
-  foreach (QChar character, string)
+  for (const auto &character : string)
   {
     // Obtain the current (x, y) cursor position
     int cursorX = m_cursorPosition.x();
@@ -981,8 +1033,7 @@ void Widgets::Terminal::removeStringFromCursor(const Direction direction,
 /**
  * @brief Initializes the terminal's data buffer.
  *
- * Clears the existing data buffer and reserves memory for 1024 * 100 lines of
- * terminal content.
+ * Clears the existing data buffer and reservers memory for the scrollback.
  *
  * This function is typically used to reset the terminal state, ensuring
  * efficient memory management for upcoming operations.
@@ -991,7 +1042,8 @@ void Widgets::Terminal::initBuffer()
 {
   m_data.clear();
   m_data.squeeze();
-  m_data.reserve(1024 * 100);
+  m_scrollOffsetY = 0;
+  m_data.reserve(MAX_LINES);
 }
 
 /**
@@ -1029,7 +1081,6 @@ void Widgets::Terminal::processText(const QChar &byte, QString &text)
   {
     appendString(text);
     text.clear();
-    m_data.append("");
     setCursorPosition(0, m_cursorPosition.y() + 1);
   }
 
@@ -1256,11 +1307,12 @@ void Widgets::Terminal::processResetFont(const QChar &byte, QString &text)
  *
  * @see cursorMoved()
  */
-void Widgets::Terminal::setCursorPosition(const QPoint position)
+void Widgets::Terminal::setCursorPosition(const QPoint &position)
 {
-  if (m_cursorPosition != position)
+  const QPoint clamped(position.x(), qBound(0, position.y(), MAX_LINES));
+  if (m_cursorPosition != clamped)
   {
-    m_cursorPosition = position;
+    m_cursorPosition = clamped;
     Q_EMIT cursorMoved();
   }
 }
@@ -1306,28 +1358,24 @@ void Widgets::Terminal::setCursorPosition(const int x, const int y)
  *
  * @see lineCount(), m_data
  */
-void Widgets::Terminal::replaceData(qsizetype x, qsizetype y, QChar byte)
+void Widgets::Terminal::replaceData(int x, int y, const QChar &byte)
 {
-  // Make sure the line at y exists in the buffer
-  if (y >= lineCount())
+  // Ensure the line exists
+  if (y >= m_data.size())
     m_data.resize(y + 1);
 
   // Get reference to current line
-  QString &currentLine = m_data[y];
+  QString &line = m_data[y];
 
-  // Ensure the current line is long enough to hold the character at x
-  if (x > currentLine.size())
-    currentLine = currentLine.leftJustified(x, ' ');
+  // Pad line to x if needed
+  if (x > line.size())
+    line = line.leftJustified(x, ' ');
 
-  // Ensure that byte is a printable character
-  if (!byte.isPrint())
-    byte = '.';
-
-  // Replace or insert the character at the cursor position
-  if (x < currentLine.size())
-    currentLine[x] = byte;
-  else
-    currentLine.append(byte);
+  // Set or append the printable character
+  if (x >= 0 && x < line.size())
+    line[x] = byte.isPrint() ? byte : '.';
+  else if (x >= 0)
+    line.append(byte.isPrint() ? byte : '.');
 }
 
 /**
@@ -1348,7 +1396,7 @@ void Widgets::Terminal::replaceData(qsizetype x, qsizetype y, QChar byte)
  * 3. `c` is neither a letter nor a number, meaning punctuation marks, symbols,
  *    and other non-alphanumeric characters will end the selection.
  */
-bool Widgets::Terminal::shouldEndSelection(const QChar c)
+bool Widgets::Terminal::shouldEndSelection(const QChar &c)
 {
   bool end = false;
   end |= c.isSpace();
@@ -1492,6 +1540,8 @@ void Widgets::Terminal::mousePressEvent(QMouseEvent *event)
     m_selectionStart = m_selectionStartCursor;
     m_selectionEnd = m_selectionStartCursor;
     m_stateChanged = true;
+
+    forceActiveFocus();
     Q_EMIT selectionChanged();
   }
 }
